-------------------------------------------------------------------------------
-- $Id: gpio_core.vhd,v 1.2 2001/11/11 21:47:40 tise Exp $
-------------------------------------------------------------------------------
-- GPIO_mod_Core - entity/architecture pair
-------------------------------------------------------------------------------
--
--                  ****************************
--                  ** Copyright Xilinx, Inc. **
--                  ** All rights reserved.   **
--                  ****************************
--
-------------------------------------------------------------------------------
-- Filename:        gpio_core.vhd
-- Version:         v1.00a
-- Description:     General Purpose Input/Output (core logic) for OPB bus
--
-------------------------------------------------------------------------------
-- Structure:
--
--              gpio_core.vhd
-------------------------------------------------------------------------------
-- Author:      BLT
-- History:
--  BLT             07-5-2001      -- First version
-- ^^^^^^
--      First version of General Purpose I/O.
-- ~~~~~~
-------------------------------------------------------------------------------
-- Naming Conventions:
--      active low signals:                     "*_n"
--      clock signals:                          "clk", "clk_div#", "clk_#x"
--      reset signals:                          "rst", "rst_n"
--      generics:                               "C_*"
--      user defined types:                     "*_TYPE"
--      state machine next state:               "*_ns"
--      state machine current state:            "*_cs"
--      combinatorial signals:                  "*_cmb"
--      pipelined or register delay signals:    "*_d#"
--      counter signals:                        "*cnt*"
--      clock enable signals:                   "*_ce"
--      internal version of output port         "*_i"
--      device pins:                            "*_pin"
--      ports:                                  - Names begin with Uppercase
--      processes:                              "*_PROCESS"
--      component instantiations:               "<ENTITY_>I_<#|FUNC>
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
library Unisim;
use Unisim.all;
library Common_v1_00_a;
use Common_v1_00_a.all;
use Common_v1_00_a.Common_Types.all;

entity GPIO_MOD_CORE is
  generic (
    C_DW         : integer := 32;
    C_AW         : integer := 32;
    C_GPIO_WIDTH : integer := 32;
    C_ALL_INPUTS : boolean := false
    );
  port (
    OPB_Clk : in std_logic;
    OPB_Rst : in std_logic;

    -- OPB signals
    OPB_ABus_Reg    : in  std_logic_vector(0 to C_AW-1);
    OPB_BE_Reg      : in  std_logic_vector(0 to C_DW/8-1);
    OPB_DBus_Reg    : in  std_logic_vector(0 to C_GPIO_WIDTH-1);
    OPB_RNW_Reg     : in  std_logic;
    OPB_select_Reg  : in  std_logic;
    OPB_seqAddr_Reg : in  std_logic;
    GPIO_DBus       : out std_logic_vector(0 to C_DW-1);
    GPIO_errAck     : out std_logic;
    GPIO_retry      : out std_logic;
    GPIO_toutSup    : out std_logic;
    GPIO_xferAck    : out std_logic;

    GPIO_IO_I :  in std_logic_vector(0 to C_GPIO_WIDTH-1);
    GPIO_IO_O : out std_logic_vector(0 to C_GPIO_WIDTH-1);
    --GPIO_IO_T : out std_logic_vector(0 to C_GPIO_WIDTH-1);

    -- Peripheral Select address decode
    GPIO_Select : in std_logic
    );
end entity GPIO_MOD_CORE;

-------------------------------------------------------------------------------
-- Architecture section
-------------------------------------------------------------------------------

architecture IMP of GPIO_MOD_CORE is

  component FDRS is
    port (
      Q : out std_logic;
      C : in  std_logic;
      D : in  std_logic;
      R : in  std_logic;
      S : in  std_logic
      );
  end component FDRS;
  component FDR is
    port (
      Q : out std_logic;
      C : in  std_logic;
      D : in  std_logic;
      R : in  std_logic
      );
  end component FDR;
  component FDRE is
    port (
      Q  : out std_logic;
      C  : in  std_logic;
      D  : in  std_logic;
      CE : in  std_logic;
      R  : in  std_logic
      );
  end component FDRE;

  signal gpio_Data_Select : std_logic;
  signal gpio_OE_Select   : std_logic;
  signal opb_Read_Reg_Rst : std_logic;
  signal opb_Read_Reg_In  : std_logic_vector(0 to C_GPIO_WIDTH-1);
  signal opb_Read_Reg_CE  : std_logic_vector(0 to C_GPIO_WIDTH-1);
  signal gpio_Data_Out    : std_logic_vector(0 to C_GPIO_WIDTH-1);
  signal gpio_Data_In     : std_logic_vector(0 to C_GPIO_WIDTH-1);
  signal gpio_OE          : std_logic_vector(0 to C_GPIO_WIDTH-1);
  signal iGPIO_xferAck    : std_logic;
  signal gpio_xferAck_Reg : std_logic;

begin  -- architecture IMP

  REG_SELECT_PROCESS : process (GPIO_Select, OPB_ABus_Reg) is
  begin
    gpio_Data_Select <= '0';
    gpio_OE_Select   <= '0';
    if GPIO_Select = '1' then
      case OPB_ABus_Reg(29) is          -- bit A29
        when '0'    => gpio_Data_Select <= '1';
        when '1'    => gpio_OE_Select   <= '1';
        when others => null;
      end case;
    end if;
  end process REG_SELECT_PROCESS;

  opb_Read_Reg_Rst <= iGPIO_xferAck or gpio_xferAck_Reg or (not GPIO_Select) or
                      (GPIO_Select and not OPB_RNW_Reg);

  READ_REG_CE_PROCESS : process(OPB_BE_Reg) is
  begin
    for i in 0 to C_GPIO_WIDTH-1 loop
      opb_Read_Reg_CE(i) <= OPB_BE_Reg((i-C_GPIO_WIDTH+C_DW)/8);
    end loop;
  end process READ_REG_CE_PROCESS;

  READ_REG_GEN : for i in 0 to C_GPIO_WIDTH-1 generate
    READ_REG_FF_I : FDRE
      port map (
        Q  => GPIO_DBus(i-C_GPIO_WIDTH+C_DW),  -- [out]
        C  => OPB_Clk,                         -- [in]
        D  => opb_Read_Reg_In(i),              -- [in]
        R  => opb_Read_Reg_Rst,                -- [in]
        CE => opb_Read_Reg_CE(i)               -- [in]
        );
  end generate READ_REG_GEN;

  TIE_DBUS_GENERATE : if C_DW > C_GPIO_WIDTH generate
    GPIO_DBus(0 to C_DW-C_GPIO_WIDTH-1) <= (others => '0');
  end generate TIE_DBUS_GENERATE;

  XFER_ACK_PROCESS : process (OPB_Clk,OPB_Rst) is
  begin
    if (OPB_Rst = '1') then
      iGPIO_xferAck <= '0';
    elsif (OPB_Clk'event and OPB_Clk = '1') then
      iGPIO_xferAck <= GPIO_Select and not gpio_xferAck_Reg;
      if iGPIO_xferAck = '1' then
        iGPIO_xferAck <= '0';
      end if;
    end if;
  end process XFER_ACK_PROCESS;

  DELAYED_XFER_ACK_PROCESS : process (OPB_Clk, OPB_Rst ) is
  begin
    if (OPB_Rst = '1') then
      gpio_xferAck_Reg <= '0';
    elsif (OPB_Clk'event and OPB_Clk = '1') then
      gpio_xferAck_Reg <= iGPIO_xferAck;
    end if;
  end process DELAYED_XFER_ACK_PROCESS;

  GPIO_xferAck <= iGPIO_xferAck;

-------------------------------------------------------------------------------
-- Registers
-------------------------------------------------------------------------------

  GPIO_INDATA_PROCESS : process(OPB_Clk) is
  begin
    if OPB_Clk = '1' and OPB_Clk'event then
      gpio_Data_In <= GPIO_IO_I;
    end if;
  end process GPIO_INDATA_PROCESS;


 -- Only_Inputs: if (C_ALL_INPUTS) generate
 --   opb_Read_Reg_In <= gpio_Data_In;
 --   GPIO_IO_T <= (others => '1');       -- All output three-stated
 -- end generate Only_Inputs;

  Inputs_And_Outputs : if not C_ALL_INPUTS generate

    READ_MUX_PROCESS : process (gpio_Data_In, gpio_OE, gpio_Data_Select,
                                gpio_OE_Select) is
    begin
      opb_Read_Reg_In <= (others => '0');
      if gpio_Data_Select = '1' then
        opb_Read_Reg_In <= gpio_Data_In;
      elsif gpio_OE_Select = '1' then
        opb_Read_Reg_In <= gpio_OE;
      end if;
    end process READ_MUX_PROCESS;

    GPIO_OUTDATA_PROCESS : process(OPB_Clk) is
    begin
      if OPB_Clk = '1' and OPB_Clk'event then
        if gpio_Data_Select = '1' and OPB_RNW_Reg = '0' then
          for i in 0 to C_GPIO_WIDTH-1 loop
            if opb_Read_Reg_CE(i) = '1' then
              gpio_Data_Out(i) <= OPB_DBus_Reg(i);
            else
              gpio_Data_Out(i) <= gpio_Data_Out(i);
            end if;
          end loop;
        end if;
      end if;
    end process GPIO_OUTDATA_PROCESS;

    GPIO_OE_PROCESS : process(OPB_Clk, OPB_Rst) is
    begin
      if (OPB_Rst = '1') then
        gpio_OE <= (others => '1');
      elsif OPB_Clk = '1' and OPB_Clk'event then
        if gpio_OE_Select = '1' and OPB_RNW_Reg = '0' then
          for i in 0 to C_GPIO_WIDTH-1 loop
            if opb_Read_Reg_CE(i) = '1' then
              gpio_OE(i) <= OPB_DBus_Reg(i);
            else
              gpio_OE(i) <= gpio_OE(i);
            end if;
          end loop;
        end if;
      end if;
    end process GPIO_OE_PROCESS;

    GPIO_IO_O <= gpio_Data_Out;
    --GPIO_IO_T <= gpio_OE;

  end generate Inputs_And_Outputs;

  GPIO_errAck  <= '0';
  GPIO_retry   <= '0';
  GPIO_toutSup <= '0';

end architecture IMP;

